package org.eso.dfs.io;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.nio.BufferUnderflowException;
import java.nio.CharBuffer;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.util.zip.GZIPInputStream;

/**
 * @author nkornwei
 *
 */
public class MemMappedFileInputStream extends InputStream {

    private MappedByteBuffer buffer;
    private FileChannel fc;

    public MemMappedFileInputStream(File f) throws IOException {
        // Open the file and then get a channel from the stream
        FileInputStream fis = new FileInputStream(f);
        fc = fis.getChannel();

        // Get the file's size and then map it into memory
        int sz = (int) fc.size();
        MappedByteBuffer b = fc.map(FileChannel.MapMode.READ_ONLY, 0, sz);
        buffer = fc.map(FileChannel.MapMode.READ_ONLY, 0, sz);
    }

    /**
     * @see java.io.InputStream#read(byte[], int, int)
     */
    public int read(byte[] arg0, int arg1, int arg2) throws IOException {
        int i;
        if (buffer.hasRemaining()) {
            i = 0;
            do {
                if (i == arg2) {
                    break;
                }
                arg0[i + arg1] = buffer.get();
                ++i;
            } while (buffer.hasRemaining());
        } else {
            i = -1;
        }
        return i;
    }

    /**
     * @see java.io.InputStream#close()
     */
    public void close() throws IOException {
        fc.close();
    }

    /**
     * 
     * @see java.io.InputStream#mark(int)
     */
    public void mark(int arg0) {
        // do nothing
    }

    /**
     * @see java.io.InputStream#markSupported()
     */
    public boolean markSupported() {
        return false;
    }

    /**
     * 
     * @see java.io.InputStream#read()
     */
    public int read() throws IOException {
        if (buffer.hasRemaining()) {
            // mask off all but the first 8 bits
            return buffer.get() & 0xFF;
        } else {
            return -1;
        }
    }

    /**
     * 
     * @see java.io.InputStream#reset()
     */
    public void reset() throws IOException {
        buffer.reset();
    }

    /**
     * 
     * @see java.io.InputStream#skip(long)
     */
    public long skip(long arg0) throws IOException {
        buffer.position(buffer.position() + (int) arg0);
        return buffer.position();
    }

    /**
     * @see java.io.InputStream#available()
     */
    public int available() throws IOException {
        System.out.println("AVAILABLE");
        return buffer.remaining();
    }

    /**
     * @see java.io.InputStream#read(byte)
     */
    public int read(byte[] arg0) throws IOException {
        return read(arg0, 0, arg0.length);
    }

    public static void main(String[] args) {
        if (args.length == 0) {
            System.out.println("Please list files on command line.");
        } else {
            for (int i = 0; i < args.length; i++) {
                File f = new File(args[i]);
                try {

                    InputStream r = new InputStreamPrinter(new MemMappedFileInputStream(f));
                    if (args[i].endsWith(".gz")) {
                        r = new GZIPInputStream(r);
                    }

                    while (true) {
                        int c = r.read();
                        if (c == -1) {
                            break;
                        }
                        System.out.print((char) c);
                    }

                } catch (IOException e) {
                    e.printStackTrace();
                    System.out.println("Error reading file: " + e.toString());
                }
            }
        }
    }
}

class InputStreamPrinter extends InputStream {

    InputStream orig;

    public InputStreamPrinter(InputStream o) {
        orig = o;
    }
    /**
     * @see java.io.InputStream#available()
     */
    public int available() throws IOException {
        System.out.println("1");
        return orig.available();
    }

    /**
     * @see java.io.InputStream#close()
     */
    public void close() throws IOException {
        System.out.println("2");
        orig.close();
    }

    /**
     * @see java.io.InputStream#mark(int)
     */
    public synchronized void mark(int arg0) {
        System.out.println("3");
        orig.mark(arg0);
    }

    /**
     * @see java.io.InputStream#markSupported()
     */
    public boolean markSupported() {
        System.out.println("4");
        return orig.markSupported();
    }

    /**
     * @see java.io.InputStream#read()
     */
    public int read() throws IOException {
        System.out.println("5");
        return orig.read();
    }

    /**
     * @see java.io.InputStream#read(byte, int, int)
     */
    public int read(byte[] arg0, int arg1, int arg2) throws IOException {
        System.out.println("6");
        return orig.read(arg0, arg1, arg2);
    }

    /**
     * @see java.io.InputStream#read(byte)
     */
    public int read(byte[] arg0) throws IOException {
        System.out.println("7");
        return orig.read(arg0);
    }

    /**
     * @see java.io.InputStream#reset()
     */
    public synchronized void reset() throws IOException {
        System.out.println("8");
        orig.reset();
    }

    /**
     * @see java.io.InputStream#skip(long)
     */
    public long skip(long arg0) throws IOException {
        System.out.println("9");
        return orig.skip(arg0);
    }

}
