/**
 * @author N. Kornweibel
 * @version 1.0
 * A JTree widget for displaying a DFSFileTreeModel object.
 * The widget supports Drag Drop using the FileListTransferHandler.
 * The widegt does nothing with drop event, it simply passes it to
 * all registered FileListImportListeners. Theses are added with
 * addFileListImportListener().
 * 
 * TODO: DFSTreeTable, ResultsTreeTable and DBTreeTable all have 
 * a CustomListener - these should be extracted out to a common type.
 * 
 * 
 */

package org.eso.gasgano.gui;

import java.awt.Dimension;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import javax.swing.JComponent;
import javax.swing.table.DefaultTableColumnModel;
import javax.swing.table.TableColumn;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import org.eso.dfs.datatransfer.FileListEvent;
import org.eso.dfs.datatransfer.FileListImportListener;
import org.eso.dfs.datatransfer.FileListTransferHandler;
import org.eso.gasgano.datamodel.filesystem.DFSFile;
import org.eso.gasgano.datamodel.gui.KeywordContainerOBComponent;
import org.eso.gasgano.datamodel.gui.OBComponent;
import org.eso.gasgano.properties.GasProp;

public class DFSFileTreeTable
    extends JTreeTable
    implements PropertyChangeListener {

    private Vector selectionVector = new Vector();
    private List importListeners = new ArrayList();

    private String selectedDir = null;

    public DFSFileTreeTable(DFSFileTreeTableModel model) {
        super(model);
        getTree().getSelectionModel().setSelectionMode(
            TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
        getTree().setCellRenderer(new DFSFileTreeCellRenderer());
        setIntercellSpacing(new Dimension(5, 1));
        setupDataTransfer();
    }

    /**
     * setup data transfer (drag&drop)
     * This feature is enabled by default for this widget.
     * null is returned from exportFileList() when anything but
     * and OBComponent is dragged.
     */
    private void setupDataTransfer() {
        setDragEnabled(true);
        setTransferHandler(new FileListTransferHandler(false, true) {
            protected List exportFileList(JComponent c) {
                Iterator i = getSelectedFiles().iterator();
                List rc = new ArrayList();
                while (i.hasNext()) {
                    String fileName = (String) i.next();
                    // System.out.println("Dragging:<"+fileName+">");
                    rc.add(new File(fileName));
                }
                if (rc.isEmpty()) {
                    rc = null;
                }
                return rc;
            }
            protected void importFileList(JComponent c, List fileList) {
                fireListListImportEvent(new FileListEvent(fileList));
            }
        });
    }

    /**
     * Add a FileList import listener to receive notification of when
     * a file list drop has occurred.
     * @param l
     */
    public void addFileListImportListener(FileListImportListener l) {
        removeFileListImportListener(l);
        importListeners.add(l);
    }

    /**
     * Remove a FileList import listener to receive notification of when
     * a file list drop has occurred.
     * @param l
     */
    public void removeFileListImportListener(FileListImportListener l) {
        importListeners.remove(l);
    }

    /**
     * Inform all listeners that a file list drop event has occurred.
     * @param l
     */
    private void fireListListImportEvent(FileListEvent event) {
        Iterator i = importListeners.iterator();
        while (i.hasNext()) {
            ((FileListImportListener) i.next()).importFileListEvent(event);
        }
    }

    /**
     * we are interested in user resizing the columns, so that
     * the preferences get updated automatically
     */

    public void propertyChange(PropertyChangeEvent pce) {
        if (pce.getPropertyName().equals("columWidth")) {
            int value = ((Integer) pce.getNewValue()).intValue();
            String name =
                (String) ((TableColumn) pce.getSource()).getHeaderValue();
            if (!name.equals("File")) {
                GasProp.columnSizeChanged(0, name, value);
            }
        } else {
            //System.out.println("PROPERTY CHANGED" + pce.getPropertyName());
        }
    }

    /**
     * adds a listener to the size of the various columns
     */

    public void registerColumnListener() {
        for (Enumeration e = getColumnModel().getColumns();
            e.hasMoreElements();
            ) {
            ((TableColumn) e.nextElement()).addPropertyChangeListener(this);
        }
    }

    public void setColumnWidths() {
        DefaultTableColumnModel tcm =
            (DefaultTableColumnModel) getColumnModel();
        Vector columnName, columnSize = null;
        TableColumn tCol = tcm.getColumn(0);
        if (GasProp
            .getProperties()
            .getProperty("SHORT_FILENAME")
            .toString()
            .equals("true")) {
            tCol.setMinWidth(370);
        } else {
            tCol.setMinWidth(450);
        }
        columnName = GasProp.getProperties().fsDisplayColumns;
        columnSize = GasProp.getProperties().fsDisplayColumnsSize;
        for (Enumeration e = getColumnModel().getColumns();
            e.hasMoreElements();
            ) {
            tCol = (TableColumn) e.nextElement();
            int pos = columnName.indexOf((String) (tCol.getHeaderValue()));
            // note that columnNames and sizes may not match up...so check
            // for boundary overflow.
            if (pos != -1 && pos < columnSize.size()) {
                int width = 0;
                if (columnSize.elementAt(pos) instanceof String) {
                    try {
                        width =
                            Integer.parseInt(
                                (String) columnSize.elementAt(pos));
                    } catch (NumberFormatException nfe) {

                    }
                } else if (columnSize.elementAt(pos) instanceof Integer) {
                    width = ((Integer) (columnSize.elementAt(pos))).intValue();
                }
                if (width != 0) {
                    //System.out.println("Setting "+tCol.getHeaderValue()+" to "+width);
                    tCol.setWidth(width);
                }
            }
        }

        // tCol.setPreferredWidth(450);
    }

    /**
     * @return Return a Vector of String of paths names of the 
     * selected files - or an empty Vector is nothing is selected.
     */
    public Vector getSelectedFiles() {
        Vector rc = selectionVector;
        rc.removeAllElements();
        TreePath paths[] = getTree().getSelectionPaths();

        if (paths != null) {
            for (int i = 0; i < paths.length; i++) {
                Object obj = paths[i].getLastPathComponent();
                if (obj instanceof OBComponent) {
                    OBComponent obc = (OBComponent) obj;
                    DFSFile dFile = obc.getSourceFile();
                    rc.addElement(dFile.getPathString());
                }
            }
        }
        return rc;
    }

    /**
     * Get the directory path of the last selected directory.
     * If multiple directories are associated with the selection, null is returned.
     * @return String the single directory assocaited with all selections.
     */
    public String getSelectedDir() {
        String rc = null;
        TreePath paths[] = getTree().getSelectionPaths();
        if (paths != null) {
            DFSFileTreeTableModel model =
                (DFSFileTreeTableModel) tree.getModel();
            for (int i = 0; i < paths.length; i++) {
                Object obj = paths[i].getLastPathComponent();
                if (!model.isLeaf(obj)) {
                    String dir = model.getDirectory(obj);
                    if (rc == null) {
                        rc = dir;
                    } else if (!rc.equals(dir)) {
                        // multiple directories in selection, return null
                        rc = null;
                        break;
                    }
                }
            }
        }
        return rc;
    }

    /**
     * Get all selected OBComponents.
     * @return Vector of OBComponent objects.
     */
    public Vector getSelectedOBComponents() {
        Vector rc = selectionVector;
        rc.removeAllElements();
        TreePath paths[] = getTree().getSelectionPaths();

        if (paths != null) {
            for (int i = 0; i < paths.length; i++) {
                Object obj = paths[i].getLastPathComponent();
                if (obj instanceof OBComponent) {
                    rc.addElement(obj);
                }
            }
        }
        return rc;
    }

    /**
     * Get the instrument associated with the selected paths, or null
     * if there is nothing selected or selection is associated with
     * multiple instruments.
     * @return String the name of the instrument associated with the selection.
     */
    public String getSelectedInstrument() {
        Vector selVect = getSelectedOBComponents();
        String rc = null;
        int c = 0;
        for (int i = 0; i < selVect.size(); i++) {
            if (selVect.elementAt(i) instanceof KeywordContainerOBComponent) {
                KeywordContainerOBComponent ffoc =
                    (KeywordContainerOBComponent) selVect.elementAt(i);
                if (c == 0) {
                    rc = ffoc.getInstrument();
                    c++;
                } else {
                    c++;
                    if (ffoc.getInstrument() != null
                        && !rc.equals(ffoc.getInstrument())) {
                        rc = null;
                        return rc;
                    }
                }
            }
        }
        return rc;
    }
    /*
        public class CustomListener extends MouseInputAdapter {
            Toolkit toolkit = null;
            Container contentPane = null;
            JFrame display = null;
            int yOffset = 0;
    
            public CustomListener(
                Container contentPane,
                JFrame display,
                int yOffset) {
                toolkit = Toolkit.getDefaultToolkit();
                this.contentPane = contentPane;
                this.display = display;
                this.yOffset = yOffset;
            }
    
            public void mouseMoved(MouseEvent e) {
                System.out.println("Moved  ");
                redispatchMouseEvent(e);
            }
    
            public void mouseDragged(MouseEvent e) {
                System.out.println("Dragged ");
                redispatchMouseEvent(e);
            }
    
            public void mouseClicked(MouseEvent e) {
                System.out.println("Clicked ");
                redispatchMouseEvent(e);
            }
    
            public void mouseEntered(MouseEvent e) {
                System.out.println("Entered ");
                redispatchMouseEvent(e);
            }
    
            public void mouseExited(MouseEvent e) {
                System.out.println("Exited ");
                redispatchMouseEvent(e);
            }
    
            public void mousePressed(MouseEvent e) {
                System.out.println("Pressed ");
                redispatchMouseEvent(e);
            }
    
            public void mouseReleased(MouseEvent e) {
                System.out.println("Released");
                redispatchMouseEvent(e);
            }
    
            void redispatchMouseEvent(MouseEvent e) {
                Component component = null;
                Point treeTablePoint = e.getPoint();
                int eventID = e.getID();
                int modifiers = e.getModifiers();
    
                treeTablePoint.y += yOffset;
    
                component =
                    SwingUtilities.getDeepestComponentAt(
                        contentPane,
                        treeTablePoint.x,
                        treeTablePoint.y);
    
                if (component != null) {
                    //System.out.println(component.getClass().getName());
                    Point componentPoint =
                        SwingUtilities.convertPoint(
                            contentPane,
                            treeTablePoint,
                            component);
    
                    if (component instanceof DFSFileTreeTable
                        && eventID == MouseEvent.MOUSE_PRESSED) {
                        JTree tree = getTree();
                        ListToTreeSelectionModelWrapper selectionModel =
                            (ListToTreeSelectionModelWrapper) tree
                                .getSelectionModel();
    
                        // Java 1.1.8 code:
                        // int row = componentPoint.y / (tree.getRowHeight() + 1);
                        // following works better for Java 1.4
                        int row = componentPoint.y / tree.getRowHeight();
    
                        DFSFileTreeTableModel model =
                            (DFSFileTreeTableModel) tree.getModel();
    
                        // it is not a leaf
                        if (!tree
                            .getModel()
                            .isLeaf(
                                tree.getPathForRow(row).getLastPathComponent())) {
                            Cursor c = display.getCursor();
                            display.setCursor(new Cursor(Cursor.WAIT_CURSOR));
                            Toolkit toolkit = Toolkit.getDefaultToolkit();
    
                            // it's  a normal single click
                            if (modifiers == InputEvent.BUTTON1_MASK) {
                                if (tree.isCollapsed(row)) {
                                    tree.expandPath(tree.getPathForRow(row));
                                } else {
                                    tree.collapsePath(tree.getPathForRow(row));
                                }
                                display.setCursor(c);
                                return;
                            }
                            if (modifiers == InputEvent.BUTTON2_MASK) {
                                Object parent =
                                    tree.getPathForRow(row).getLastPathComponent();
                                selectedDir = model.getDirectory(parent);
                                toolkit.beep();
                                selectionModel.clearSelection();
                                display.setCursor(c);
                                return;
                            }
                            if ((modifiers & InputEvent.BUTTON3_MASK) != 0) {
                                // it's a right mouse button click
                                Object parent =
                                    tree.getPathForRow(row).getLastPathComponent();
                                //int numChild = tree.getModel().getChildCount( parent );
                                int i = 0;
                                Vector allLeaves = model.getAllTheLeaves(parent);
                                TreePath[] kids = new TreePath[allLeaves.size()];
    
                                //Runtime r = Runtime.getRuntime();
                                //r.gc();
    
                                for (Enumeration ee = allLeaves.elements();
                                    ee.hasMoreElements();
                                    ) {
                                    // this is a very unpleasant particularization
                                    // which I would have liked to avoid.....
                                    Object tmpNode = ee.nextElement();
                                    kids[i] =
                                        new TreePath(
                                            model.getPathToRoot(
                                                (OBComponent) tmpNode));
                                    i++;
                                }
    
                                if (e.isControlDown()) {
                                    // try to determine if they are all already selected
                                    boolean allSelected = true;
                                    for (int l = 0; l < kids.length; l++) {
                                        if (!selectionModel
                                            .isPathSelected(kids[l])) {
                                            allSelected = false;
                                            break;
                                        }
                                    }
                                    if (allSelected) {
                                        selectionModel.removeSelectionPaths(kids);
                                    } else {
                                        selectionModel.addSelectionPaths(kids);
                                    }
                                } else {
                                    selectionModel.setSelectionPaths(kids);
                                }
                                display.setCursor(c);
                                // how to cope with ah buggy garbage collector:
                                for (int l = 0; l < kids.length; l++) {
                                    kids[l] = null;
                                }
                                kids = null;
    
                                return;
                            }
                        }
                    }
                    component.dispatchEvent(
                        new MouseEvent(
                            component,
                            e.getID(),
                            e.getWhen(),
                            e.getModifiers(),
                            componentPoint.x,
                            componentPoint.y,
                            e.getClickCount(),
                            e.isPopupTrigger()));
                }
    
            }
        }
    */
}
// end of file DFSFileTreeTable.java
