package org.eso.ohs.core.docview.datatrans;

import org.apache.log4j.Logger;


/**
 * This is a singleton factory class for ObjectAdaptor objects.
 * It can be subclassed by the application and an instance of the
 * subclass installed, thus allowing the application to tailor
 * the mapping of ObjectAdaptor to property type.
 * <p>
 * It is used by DocView.setObjectProperty() but an application
 * may find ObjectAdaptors more generally useful, since they hide
 * the machinery of extracting property values from a business
 * object.
 * 
 * <p>
 * $Id: AdaptorFactory.java,v 1.3 2004/04/08 12:09:51 tcanavan Exp $
 * <p>
 * @author Bob Kemp
 * 
 * @see org.eso.ohs.core.docview.gui.DocView
 */
public class AdaptorFactory {

   /** Standard log channel */
   private static Logger stdlog_ =
       Logger.getLogger( AdaptorFactory.class );


    /** Singleton instance of this class.  Can be replaced using setCfg() */
    protected static AdaptorFactory instance_ = new AdaptorFactory();


    /** Null constructor */
    protected AdaptorFactory() {
    }

    /**
     * Returns the singleton instance of this class (or a subclass).
     * It is expected that applications will usually use setCfg()
     * to install an application-specific subclass of AdaptorFactory.
     * 
     * @return	singleton instance of AdaptorFactory
     */
    public static AdaptorFactory getInstance() {
       return instance_; 
    }

    /**
     * Install a new instance of AdaptorFactory.
     * 
     * @param cfg	instance of (a subclass of) AdaptorFactory
     */
    public static void setInstance(AdaptorFactory instance) {
       instance_ = instance;
    }


    /**
     * Returns an ObjectAdaptor that links the value of the named property
     * of the Java object to the String-valued setValue()/getValue()
     * methods of the ObjectAdaptor.  In other words, setValue() on
     * the adaptor will set the value of the given property.
     * The class represented by <code>cls</code> must implement
     * the DocViewObject interface.
     * 
     * @param cls	java.lang.Class object for the class to be linked
     * @param property	Name of a property of the cls class
     * 
     * @return	an ObjectAdaptor for the property and class
     * 
     * @exception IllegalArgumentException
     * 		thrown if there is no such public property in the class.
     */
    public static ObjectAdaptor getAdaptor(Class cls, String property) {
	return getInstance().getAdaptorImpl(cls, property);
    }


    /**
     * Returns an ObjectAdaptor that links the value of the named property
     * of the Java object to the String-valued setValue()/getValue()
     * methods of the ObjectAdaptor.  In other words, setValue() on
     * the adaptor will set the value of the given property.
     * 
     * @param dvObj	object to be linked
     * @param property	Name of a property of the cls class
     * @param readWrite	True if property is writeable, false if read-only.
     * 
     * @return	an ObjectAdaptor for the property and class
     * 
     * @exception IllegalArgumentException
     * 		thrown if there is no such public property in the class.
     */
    public static ObjectAdaptor getAdaptor(Object dvObj,String property,
						boolean readWrite) {
	ObjectAdaptor oa = AdaptorFactory.getAdaptor(
					dvObj.getClass(),
					property
				    );

	oa.setReadWrite(readWrite);
	oa.setObject(dvObj);
	oa.setProperty(property);

	return oa;
    }


    /**
     * Returns an ObjectAdaptor that links the value of the named property
     * of the Java object to the String-valued setValue()/getValue()
     * methods of the ObjectAdaptor.  In other words, setValue() on
     * the adaptor will set the value of the given property.
     * 
     * @param dvObj	object to be linked
     * @param property	Name of a property of the cls class
     * 
     * @return	an ObjectAdaptor for the property and class
     * 
     * @exception IllegalArgumentException
     * 		thrown if there is no such public property in the class.
     */
    public static ObjectAdaptor getAdaptor(Object dvObj,String property){
	return getAdaptor(dvObj, property, true);
    }


    /**
     * Internal implementation method to return an ObjectAdaptor
     * that links the value of the named property of the Java object
     * to the String-valued setValue()/getValue() methods of the
     * ObjectAdaptor.  In other words, setValue() on the adaptor
     * will set the value of the given property.  The class
     * represented by <code>cls</code> must implement the
     * object interface.
     * 
     * <p>
     * This method is intended to be overriden by application-specific
     * subclasses of AdaptorFactory.
     * 
     * @param cls	java.lang.Class object for the class to be linked
     * @param property	Name of a property of the cls class
     * 
     * @return	an ObjectAdaptor for the property and class
     * 
     * @exception IllegalArgumentException
     * 		thrown if there is no such public property in the class.
     */
    protected ObjectAdaptor getAdaptorImpl(Class cls, String property) {
	try {
	    return ObjAdaptor.getAdaptor(cls, property);
	}
	catch (NoSuchMethodException e) {
//	    stdlog_.debug("class    = " + cls);
//	    stdlog_.debug("property = " + property);
	    stdlog_.error(e);
	    throw new IllegalArgumentException(
			    "No such property: "
			    + cls.getName() + ": "
			    + property
			);
	}
    }

}

