package org.eso.ohs.core.utilities;

import java.io.File;
import java.io.FilenameFilter;


/**
 * This class returns a list of filenames from or file handles for 
 * a directory that have a particular extension.
 * 
 * @author Tim Canavan
 */

public class ExtFileList {   
    
    /**
     * List of filenames based on a file extension.
     */
    private String[] fileList_;
   
 

    /**
     * Handle to a file in this case a directory.
     */
    private File dir_;


    /**
     * Inner class which filters out the files with the neccessary 
     * extension.
     */

    private class ExtensionFilenameFilter implements FilenameFilter {
        /**
         * File extension that the filter selects on
         */
        private String extension_;

        /**
         * Constructor that excepts an extension as its file filter.
         * @param extension file extension
         */
        public ExtensionFilenameFilter(String extension) {
            extension_ = extension;
        }

        /**  
         * Decide whether or not a file should be part of the 
         * file list.
         * @param dir directory
         * @param name directory name
         */
        public boolean accept(File dir,String name) {
            if(name.endsWith(extension_))
                return true;
            return false;
        }        
    }

    ////////////////////////////////////////////////////////////////
    // Constructor(s)                                              /
    ////////////////////////////////////////////////////////////////

    /**
     * Constructs a list of Java File types given a directory and extension
     * @param directory directroy to be scanned
     * @param extension only files with these extensions to be added to the List
     * @exception IllegalArgumentException If directory is not valid
     */

    public ExtFileList(String directory,String extension) throws IllegalArgumentException {
        ExtensionFilenameFilter filter = new ExtensionFilenameFilter(extension);
        dir_ = new File(directory);
        if(!dir_.isDirectory())
            throw new IllegalArgumentException("No Such Directory " + directory);
        fileList_ = dir_.list(filter);
    
    }


    public boolean isEmpty() {
        if(fileList_.length <= 0)
            return true;
        return false;
    }

    ////////////////////////////////////////////////////////////////
    // Methods                                                     /
    ////////////////////////////////////////////////////////////////

    /**
     * Return a file handle in the list
     * @param filename filename
     */

    public File getFileHandle(String filename) {
        String absPath = dir_.getAbsolutePath() + File.separator + filename;        
        File f = new  File(absPath);       
        boolean inList = exists(filename);
        if(!f.exists() || !inList)
            return (File)null;
            //throw new IllegalArgumentException("File not in list " + filename);
        return f;
    }

    /**
     * File is part of the list
     * @param file in list
     * @return True if file is in file list else false     
     */

    public boolean exists(String filename) {
        boolean inList = false;
        for(int i=0;i<fileList_.length;i++) {
	    if(filename.equals(fileList_[i])) {
                inList = true;
            }
        }        
        return inList;
    }



    /**
     *  Retruns  a list of file names as strings
     *  @return list of filenames as strings
     */
    public String[] getFileList() {   
        return fileList_; 
    }

    /**
     * Returns a list of file handles for a given extension
     * @return an array of file handles.
     */

    public File [] getFileHandles() {
        if(isEmpty())
             return new File[0];
        File flist [] = new File[fileList_.length];
        for(int i=0;i<fileList_.length;i++) {
            flist[i] = new File(dir_.getAbsolutePath() + File.separator + fileList_[i]);
        }
        return flist;
    }
}






