package org.eso.ohs.phase2.apps.ot.gui;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Hashtable;

import javax.swing.tree.DefaultMutableTreeNode;

import org.eso.ohs.core.docview.datatrans.AdaptorFactory;
import org.eso.ohs.core.utilities.Convert;
import org.eso.ohs.dfs.CalibrationBlock;
import org.eso.ohs.dfs.ConstraintSet;
import org.eso.ohs.dfs.ObservationBlock;
import org.eso.ohs.dfs.ObservationDescription;
import org.eso.ohs.dfs.Target;
import org.eso.ohs.dfs.TimeInterval;
import org.eso.ohs.dfs.docviewdatatrans.DegreeAdaptor;
import org.eso.ohs.instruments.Instrument;
import org.eso.ohs.instruments.InstrumentConstraintConfig;
import org.eso.ohs.instruments.InstrumentList;
import org.eso.ohs.instruments.Parameter;
import org.eso.ohs.instruments.TemplateSignature;

/**
 * A Factory class for creating a set of node usign 
 * observationBlock's fields
 *
 * @author Gorgio Passera <gpassera@eso.org>
 * <br>ChangeLog: cmavros 1647/OT  Add user comments to tree
 */
public class OBTreeNodeFactory {

	private static final int tabPos = 35;
	private static float NULL_FLOAT = 0;

	private OBTreeNodeFactory() {
	}

	/**
	*Creates a set of node contaning information regarding an 
	*Observation block
	*
	*@param ob specifies an observation block
	*@return the tree root node
	*/
	public static DefaultMutableTreeNode createCB(CalibrationBlock cb) {

		DefaultMutableTreeNode root =
			new DefaultMutableTreeNode(
				cb.getObsRun().getName() + "-" + cb.getName());

		addDetails(root, cb);
		addTemplateSignature(root, cb);

		return root;
	}

	/**
	  *Creates a set of node contaning information regarding an 
	  *Observation block
	  *
	  *@param ob specifies an observation block
	  *@return the tree root node
	  */
	public static DefaultMutableTreeNode createOB(ObservationBlock ob) {

		DefaultMutableTreeNode root =
			new DefaultMutableTreeNode(
				ob.getObsRun().getName() + "-" + ob.getName());

		addDetails(root, ob);
		addTarget(root, ob.getTarget());
		try {
            addConstraintSet(root, ob);
        } catch (SecurityException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (IllegalArgumentException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (NoSuchMethodException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (IllegalAccessException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
		addTimeIntervals(root, ob.getTimeIntervals());
                addSTTimeIntervals(root, ob.getSTTimeIntervals());
		addAcquisitionTemplate(root, ob);
		addTemplateSignature(root, ob);

		return root;
	}

	/**Adds details*/
	private static void addDetails(
		DefaultMutableTreeNode root,
		CalibrationBlock ob) {

		DefaultMutableTreeNode node = new DefaultMutableTreeNode("Details ...");

		add(node, "Repository ID", ob.getId());
		add(node, "Observer comments", ob.getObsComments());
		if (ob instanceof ObservationBlock)
			add(
				node,
				"Observer QC grade",
				((ObservationBlock) ob).getObsGrade());
		add(node, "User priority", ob.getUserPriority());
		add(node, "Rank Class", ob.getObsRun().getRankClass());

		//add(node,"OPC priority","???");
		//add(node,"clone of",ob.getCloneParentId());
		add(node, "Observing mode", ob.getOd().getMode());
		add(node, "Status", ob.getStatus());
		add(node, "User ID", ob.getOwnerId());
		add(node, "User Comments", ob.getUserComments());
        Instrument instObj = InstrumentList.getInstance().getInstrument(ob.getOd().getInstrument(),ob.fetchIPVersionFromOd());
        if(instObj.hasInstComments()) {
            String comments = ob.getInstrumentComments();
            if(comments == null)
                comments="";
            add(node,instObj.getLabelInstComment(),comments);
        }
        ObservationDescription od = ob.getOd();
        if (od != null) {
            // add this info if we have an obs description.
            add(node, "Instrument", od.getInstrument());
            add(node, "IP Version", od.getIPVersion());
        }

		root.add(node);
	}

	/**Add a target*/
	private static void addTarget(DefaultMutableTreeNode root, Target target) {

		StringBuffer buf = new StringBuffer(target.getName());

		DegreeAdaptor oa = null;

		oa = (DegreeAdaptor) AdaptorFactory.getAdaptor(Target.class, "RA");

		buf.append(" " + oa.fromMetaLevelObject(new Integer(target.getRA())));

		oa =
			(DegreeAdaptor) AdaptorFactory.getAdaptor(
				Target.class,
				"Declination");

		buf.append(
			" , "
				+ oa.fromMetaLevelObject(new Integer(target.getDeclination())));

		DefaultMutableTreeNode node =
			new DefaultMutableTreeNode(buf.toString());
		add(node, "Equinox", target.getEquinox());
		add(node, "Prop. motion alpha", target.getPropRA(), NULL_FLOAT);
		add(
			node,
			"Prop. motion delta",
			target.getProperDeclination(),
			NULL_FLOAT);
		add(node, "Magnitude", target.getMag(), NULL_FLOAT);
		add(node, "Diff motion alpha", target.getDiffRA(), NULL_FLOAT);
		add(node, "Diff motion delta", target.getDiffDec(), NULL_FLOAT);
		add(node, "User Comments", target.getUserComments());

		root.add(node);
	}

	/**Adds a constraintset*/
	private static void addConstraintSet(
		DefaultMutableTreeNode root,
		ObservationBlock ob) throws SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		ConstraintSet cs = ob.getConstraintSet();

                

		if (cs != null) {
            Instrument instObj = InstrumentList.getInstance().getInstrument(ob.getOd().getInstrument(),ob.fetchIPVersionFromOd());
            InstrumentConstraintConfig [] insConfList = instObj.getCsConfList();

			DefaultMutableTreeNode node =
				new DefaultMutableTreeNode(/*"Constraint set :" + */
			cs.getName());
            for(int i=0;i<insConfList.length;i++) {
                String propName = insConfList[i].getName();
                Class cls = cs.getClass();
                Method get;
                get = cls.getMethod("get" + insConfList[i].getName(), new Class[] {
                });
                Object obje = get.invoke(cs, new Object[] {});
                if(obje == null) {
                    obje= new String("NONE");
                }
                add(node,insConfList[i].getName(),obje.toString());
            }
			root.add(node);
		}

	}

	/*Adds all time intervals*/
	private static void addTimeIntervals(
		DefaultMutableTreeNode root,
		TimeInterval[] ti) {
		if ((ti != null) && (ti.length > 0)) {
			DefaultMutableTreeNode node =
				new DefaultMutableTreeNode("Time Intervals");
			for (int i = 0; i < ti.length; i++) {
				addTimeInterval(node, ti[i]);
			}
			root.add(node);
		}
	}

	/*Adds a single time interval*/
	private static void addTimeInterval(
		DefaultMutableTreeNode root,
		TimeInterval ti) {
		//stdlog_.debug("AA");
		add(root, "Start Time",Convert.longToISODate(ti.getStartTime()));
		add(root, "End Time", Convert.longToISODate(ti.getEndTime()));
	}


	/*Adds all time intervals*/
	private static void addSTTimeIntervals(
		DefaultMutableTreeNode root,
		TimeInterval[] ti) {
		if ((ti != null) && (ti.length > 0)) {
			DefaultMutableTreeNode node =
				new DefaultMutableTreeNode("Sidereal Time Intervals");
			for (int i = 0; i < ti.length; i++) {
				addSTTimeInterval(node, ti[i]);
			}
			root.add(node);
		}
	}

	/*Adds a single time interval*/
	private static void addSTTimeInterval(
		DefaultMutableTreeNode root,
		TimeInterval ti) {
		//stdlog_.debug("AA");
		add(root, "Start Time", Convert.longToHHSS(ti.getStartTime()));
		add(root, "End Time", Convert.longToHHSS(ti.getEndTime()));
	}


	/**Adds a User Comments
	 * <br>ChangeLog: cmavros 1647/OT  Add user comments to tree
	 */
	private static void addUserComments(
		DefaultMutableTreeNode root,
		ObservationBlock ob) {
		DefaultMutableTreeNode node =
			new DefaultMutableTreeNode("User Comments");

		add(node, "User Comments", ob.getUserComments());

		root.add(node);

	}

	/**Adds template signature from an ObservationDescription*/
	private static void addTemplateSignature(
		DefaultMutableTreeNode root,
		CalibrationBlock ob) {
		TemplateSignature[] ts = ob.getOd().getSignatures();
		if (ts != null) {
			for (int i = 0; i < ts.length; i++) {

				DefaultMutableTreeNode child =
					new DefaultMutableTreeNode(ts[i].getTemplateName());
				root.add(child);
				addParamTree(child, ts[i].getParamList());
			}
		}
	}

	/**Adds Acquisition template*/
	private static void addAcquisitionTemplate(
		DefaultMutableTreeNode root,
		ObservationBlock ob) {
		TemplateSignature ts = ob.getTemplateSignature();
		if (ts != null) {
			DefaultMutableTreeNode node =
				new DefaultMutableTreeNode(/*"Acquisition Template :"+ */
			ob.getTemplateSignature().getTemplateName());
			addParamTree(node, ts.getParamList());
			root.add(node);
		}
	}

	/**
	 *Adds a child node to the parent .Child node named is composed using 
	 *the couple name-value . Child node is created only if "value" 
	 *is different from "nullValue"  
	 *
	 *@param parent a parent node
	 *@param name 
	 *@param value
	 *@param nullValue
	 */
	private static void add(
		DefaultMutableTreeNode parent,
		String name,
		long value,
		long nullValue) {

		if (value != nullValue) {
			add(parent, name, value);
		}

	}

	/**
	 *Adds a child node to the parent .Child node named is composed using 
	 *the couple name-value
	 *
	 *@param parent a parent node
	 *@param name 
	 *@param value
	 */
	private static void add(
		DefaultMutableTreeNode parent,
		String name,
		long value) {

		StringBuffer buf = new StringBuffer();
		buf.append(value);
		DefaultMutableTreeNode node =
			new DefaultMutableTreeNode(buildText(name, buf.toString()));
		parent.add(node);
	}

	private static void add(
		DefaultMutableTreeNode parent,
		String name,
		int value) {
		StringBuffer buf = new StringBuffer();
		buf.append(value);
		DefaultMutableTreeNode node =
			new DefaultMutableTreeNode(buildText(name, buf.toString()));
		parent.add(node);
	}

	private static void add(
		DefaultMutableTreeNode parent,
		String name,
		float value) {

		StringBuffer buf = new StringBuffer();
		buf.append(value);
		DefaultMutableTreeNode node =
			new DefaultMutableTreeNode(buildText(name, buf.toString()));
		parent.add(node);
	}

	private static void add(
		DefaultMutableTreeNode parent,
		String name,
		float value,
		float nullValue) {

		if (value != nullValue) {
			add(parent, name, value);
		}
	}

	private static void add(
		DefaultMutableTreeNode parent,
		String name,
		String value) {
		if (value == null)
			value = "";
		//if ( (value != null) &&  (value.length() > 0)) {
		StringBuffer buf = new StringBuffer();
		buf.append(value);
		DefaultMutableTreeNode node =
			new DefaultMutableTreeNode(buildText(name, buf.toString()));
		parent.add(node);
		//}

	}

	/**
	 *Creates a message with the following convenstion: "NAME : VALUE"
	 *@param name specifies a parameter name
	 *@param value specifies a parameter value
	 *
	 *@return  the message
	 */
	private static StringBuffer buildText(String name, String value) {
		StringBuffer buf = new StringBuffer(name + " :");
		/*
		int length = tabPos - buf.length();
		//      stdlog_.debug("buf >" + buf +"< len= " + buf.length() +
		                   "pos = " + length);
		for (int i=0; i< length ; i++) {
		buf.append(".");
		}
		//      stdlog_.debug(" buf len = " + buf.length()); 
		*/

		buf.append(" " + value);
		return buf;
	}

	/**adds a paramter tree*/
	private static void addParamTree(
		DefaultMutableTreeNode root,
		Parameter[] parameters) {

		Hashtable keyTable = new Hashtable();

		if ((parameters != null) && (parameters.length > 0)) {
			for (int i = 0; i < parameters.length; i++) {
				String name = parameters[i].getParameterName();
				// stdlog_.debug("Parameter Name " + name);		
				int pos = name.indexOf('.');
				String mainKey = name.substring(0, pos);
				String subKey = name.substring(pos + 1);
				// stdlog_.debug("Subkey " + subKey);
				DefaultMutableTreeNode node =
					(DefaultMutableTreeNode) keyTable.get(mainKey);
				if (node == null) {
					node = new DefaultMutableTreeNode(mainKey);
					root.add(node);
					keyTable.put(mainKey, node);
				}

				add(node, subKey, parameters[i].getDisplayValue());
			}

		}
	}

}
