package org.eso.ohs.phase2.apps.p2ppAppServer;

import java.io.IOException;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.FindingChart;
import org.eso.ohs.dfs.ObservingRun;
import org.eso.ohs.dfs.Readme;
import org.eso.ohs.dfs.Summary;
import org.eso.ohs.instruments.Instrument;
import org.eso.ohs.instruments.InstrumentList;
import org.eso.ohs.instruments.TemplateSignature;
import org.eso.ohs.instruments.TemplateSignatureFactory;
import org.eso.ohs.instruments.TemplateVersionInfo;
import org.eso.ohs.persistence.DirectoryNode;
import org.eso.ohs.persistence.ObjectNotFoundException;
import org.eso.ohs.persistence.StorageManager;
import org.eso.ohs.persistence.StorageManager.IdType;
import org.eso.ohs.persistence.StorageManager.MultiObjSpec;
import org.eso.ohs.persistence.StorageManager.UnpackedStorableObject;

/**
 * The storage manager class for files stored across the network via RMI.
 * A concrete subclass of StorageManager.
 * 
 * <p>
 * $Id: RmiStorageManager.java,v 1.16 2004/12/07 10:04:24 tcanavan Exp $
 * <br>
 * @author Bob Kemp
 * <br>ChangeLog: cmavros 22Apr02 1195/P2PP Validation of OBs through P2PP AppServer.
 * 
 * @see org.eso.ohs.persistence.StorageManager
 * @see org.eso.ohs.phase2.apps.p2pp.Persistence
 */

public class RmiStorageManager
	extends UnicastRemoteObject
	implements RmiManagerInterface {

	static final long serialVersionUID = -7771545424395597950L;

	private StorageManager stMgr_;

	private boolean authenticated_ = false;
    
    /**
     * Standard log channel 
     */
    private static Logger stdlog_ =
        Logger.getLogger(RmiStorageManager.class);

	public RmiStorageManager(StorageManager stMgr) throws RemoteException {
		super();

		stMgr_ = stMgr;
	}

	/**
	 * Verify that the TemplateSignature files used by the StorageManager
	 * are the same version as those specified (and used by the client)
	 *
	 * Note (from Nick Kornweibel):
	 *
	 * If a template is not found, it is added to the error list with a prefixed
	 * "- " string. This way the client can differentiate between an out of date template
	 * and a missing template. This prefix is not defined anywhere - it is hard coded into
	 * the client and server code. It really needs to be handled as an exception, which would
	 * require changing the RMI interface.
	 * The client code for this is: CheckInAction.java
	 * 
	 * @param tsList	array of version info objects for each template
	 *			used in an OB
	 * 
	 * @return a list of the names of TemplateSignatures that are out of date.
	 * 
	 * @exception ObjectIOException
	 *			if there was an I/O error while checking
	 */
	public String[] verifyTemplateSignatures(TemplateVersionInfo[] tslist)
		throws ObjectIOException, RemoteException {
		try {
			Vector errorList = new Vector();
			String[] tmplErrorList;

			for (int j = 0; j < tslist.length; j++) {
				TemplateVersionInfo templInfo = tslist[j];
                TemplateSignature ts = null;
				Instrument inst =
					InstrumentList.getInstance().getInstrument(
						templInfo.getInstrumentName(),
						templInfo.getVersion());
                
                if (inst == null) {
                    // no instrument, but it can only reported as a missing template.
                    stdlog_.error("Missing Instrument : "
                        +templInfo.getInstrumentName()+" v"+templInfo.getVersion());
                    ts = null;
                } else {
					ts = TemplateSignatureFactory.getTemplate(
						inst,
						templInfo.getTemplateName());
                }
				/**
				 * handle the issue that the template is missing.
				 */
				if (ts == null) {
					/*
					 * See not in method header for details of the "- " prefix.
					 */
					errorList.addElement("- " + templInfo.getTemplateName());
					continue;
				}
				if (ts.getVersion() != templInfo.getVersion()) {
					if (ts.getVersion() == -1
						|| templInfo.getVersion() == -1) {
						errorList.addElement(ts.getTemplateName());
						continue;
					}

					if (templInfo.getVersion() != ts.getVersion()) {
						errorList.addElement(ts.getTemplateName());
						continue;
					}
				}
			}
			tmplErrorList = new String[errorList.size()];
			errorList.copyInto(tmplErrorList);
			return tmplErrorList;
		} catch (IOException e) {
			throw new ObjectIOException(e);
		}
	}






	/**
	 * Calculates a unique id for a StorableObject.
	 *
	 * @return Unique id
	 * 
	 * @exception ObjectIOException
	 *			if the method call generated an I/O error
	 */
	public long getNewId(Class mo, long baseId) throws ObjectIOException {
		return stMgr_.getNewId(mo, baseId);
	}

	/**
	 * Get the root node of the DirectoryNode tree representing
	 * the directories found on this medium.  For example, a
	 * representation of the root of the local StorableObject
	 * storage subtree.
	 * 
	 * @return a representation of the root node
	 * 
	 * @exception ObjectIOException
	 *			if an I/O error occurred during the method's execution.
	 */
	public DirectoryNode getRoot() throws ObjectIOException {
		return stMgr_.getRoot();
	}

	/**
	 * Sets the id of the current user.
	 * 
	 * @param userId	the id of the end-user
	 * 
	 * @exception ObjectIOException
	 *			if there was an I/O error while setting the userId
	 */
	public void setUserId(int userId) throws ObjectIOException {
		authenticated_ = false;
		stMgr_.setUserId(userId);
	}

	/**
	 * Gets the id of the current user.
	 * 
	 * @return	the id of the end-user
	 */
	public int getUserId() {
		return stMgr_.getUserId();
	}

	/**
	 * Authenticates the end-user.
	 * 
	 * @param password	password for the current user
	 * 
	 * @return	true if user was successfully authenticated, false otherwise
	 * 
	 * @exception ObjectIOException
	 *			if there was an I/O error while authenticating
	 */
	public boolean loginUser(String password)
		throws ObjectIOException, RemoteException {
		try {
			authenticated_ = stMgr_.loginUser(password);
		} catch (RuntimeException e) {
			e.printStackTrace();
			authenticated_ = false;
		}
		return authenticated_;
	}

	/**
	 * Returns a DirectoryNode representing the path within the
	 * storage medium of an object with the given id and type.
	 * For media that do not support directories, the root directory
	 * may be returned.
	 * 
	 * @param id	id of the StorableObject.
	 * @param objType	type of stored StorableObject.
	 * 
	 * @return Directory in which to find the object.
	 * 
	 * @exception ObjectNotFoundException
	 *			if the object was not found.
	 */
	public DirectoryNode find(long id, Class objType)
		throws ObjectNotFoundException, ObjectIOException {
		return stMgr_.find(id, objType);
	}

	/**
	 * Returns an array of objects representing the contents of an
	 * object with the given id and type from the chosen directory.
	 * If the medium does not support directories, the <code>node</code>
	 * parameter is ignored.
	 * 
	 * @param node	directory from which the object should be loaded.
	 * @param id	id of the StorableObject.
	 * @param objType	type of StorableObject.
	 * 
	 * @return the array of objects and object references,
	 * 			as described above.
	 * 
	 * @exception ObjectNotFoundException
	 *			if the object was not found.
	 * @exception ObjectIOException
	 * 			if an I/O error occurred while reading the object
	 * 
	 * @see #write
	 * @see #update
	 */
	public UnpackedStorableObject read(
		DirectoryNode node,
		long id,
		Class objType)
		throws ObjectNotFoundException, ObjectIOException {
		try {
			return stMgr_.read(node, id, objType);
		} catch (ObjectIOException e) {
			e.printStackTrace();
			throw e;
		} catch (ObjectNotFoundException e) {
			e.printStackTrace();
			throw e;
		}
	}

	/**
	 * Writes an array of objects with the given id and type to
	 * this storage medium, in the chosen directory.  If the medium
	 * does not support directories, the <code>node</code> parameter
	 * is ignored.
	 * 
	 * @param node	directory in which the object should be saved.
	 * @param id	id under which to store the StorableObject.
	 * @param objType	type of StorableObject.
	 * @param obj	array of objects and object references to store,
	 *			as described in the section for the read method.
	 * 
	 * @exception ObjectIOException
	 * 			if an I/O error occurred while writing the object
	 * 
	 * @see #read
	 * @see #update
	 */
	public void write(
		DirectoryNode node,
		long id,
		Class objType,
		UnpackedStorableObject obj)
		throws ObjectIOException {
		stMgr_.write(node, id, objType, obj);
	}

	/**
	 * Updates an array of objects with the given id and type on
	 * this storage medium, in the chosen directory.  If the medium
	 * does not support directories, the <code>node</code> parameter
	 * is ignored.
	 * 
	 * @param node	directory in which the object should be saved.
	 * @param id	id under which to store the StorableObject.
	 * @param objType	type of StorableObject.
	 * @param obj	array of objects and object references to store,
	 *			as described in the section for the read method.
	 * 
	 * @exception ObjectIOException
	 * 			if an I/O error occurred while updating the object
	 * 
	 * @see #read
	 * @see #write
	 */
	public void update(
		DirectoryNode node,
		long id,
		Class objType,
		UnpackedStorableObject obj)
		throws ObjectIOException {
		stMgr_.update(node, id, objType, obj);
	}

	/**
	 * Deletes an object with the given directory, id and type from
	 * this storage medium.  If the medium does not support
	 * directories, the <code>node</code> parameter is ignored.
	 * 
	 * @param node	directory from which the object should be deleted.
	 * @param id	id of the StorableObject.
	 * @param objType	type of StorableObject.
	 * 
	 * @exception ObjectIOException
	 * 			if an I/O error occurred while updating the object
	 * 
	 * @see #read
	 * @see #update
	 */
	public void delete(DirectoryNode node, long id, Class objType)
		throws ObjectIOException {
		stMgr_.delete(node, id, objType);
	}

	/**
	 * Deletes an object with the given directory, id and type from
	 * this storage medium.  If the medium does not support
	 * directories, the <code>node</code> parameter is ignored.
	 * 
	 * @param node	directory from which the object should be deleted.
	 * @param id	id of the StorableObject.
	 * @param objType	type of StorableObject.
	 * 
	 * @exception ObjectIOException
	 * 			if an I/O error occurred while updating the object
	 * 
	 * @see #read
	 * @see #update
	 */
	public void move(
		long id,
		Class objType,
		DirectoryNode fromNode,
		DirectoryNode toNode)
		throws ObjectNotFoundException, ObjectIOException {
		stMgr_.move(id, objType, fromNode, toNode);
	}

	/**
	 * Create a directory on the external medium.
	 *
	 * @param node The directory to be created
	 * @param name The name of the directory to be created.
	 * 
	 * @return True on success false on error
	 * 
	 * @exception ObjectIOException
	 * 			if an I/O error occurred during the operation
	 */
	public boolean createDirectory(DirectoryNode node, long id)
		throws ObjectIOException {
		return stMgr_.createDirectory(node, id);
	}


        public FindingChart [] populateFindingCharts(long boId) throws ObjectIOException {
            return stMgr_.populateFindingCharts(boId);
        }


	/**
	 * Move a directory.
	 *
	 * @param node	directory to be moved
	 * @param toNode	the new parent directory
	 * 
	 * @return True on success false on error
	 * 
	 * @exception ObjectIOException
	 * 			if an I/O error occurred during the operation
	 * 
	 * @exception ObjectNotFoundException
	 *			if the object was not found in fromNode.
	 */
	public boolean moveDirectory(DirectoryNode fromNode, DirectoryNode toNode)
		throws ObjectNotFoundException, ObjectIOException {
		return stMgr_.moveDirectory(fromNode, toNode);
	}

	/**
	 * Delete a directory from external media
	 *      
	 * @param node	Directory to be deleted
	 * 
	 * @return True on success, false on failure
	 * 
	 * @exception ObjectIOException
	 * 			if an I/O error occurred during the operation
	 */
	public boolean deleteDirectory(DirectoryNode node)
		throws ObjectIOException {
		return stMgr_.deleteDirectory(node);
	}

	/**
	 * Returns a list of ObjectInfo objects associated with a directory.
	 * 
	 * @param node Directory to be used to build up list of objects
	 * @return An array of ObjectFileInfo[] objects 
	 * 
	 * @exception ObjectIOException
	 * 			if an I/O error occurred while updating the object
	 */
	public Summary[] listObjects(DirectoryNode node) throws ObjectIOException {
		return stMgr_.listObjects(node);
	}

	/**
	 * Do a EVM verification test for the specified objects in a single atomic operation.
	 *
	 * @param   usos    Descriptions of each StorableObject to do EVM verification
	 *
	 * @return boolean if OB passed EVM test
	 */
	public boolean verifyObject(MultiObjSpec[] moi) throws ObjectIOException {
		return stMgr_.verifyObject(moi);
	}

	/**
	 * @return the requested IP as a byte array, or an empty (size 0) array
	 * if the IP does not exist on the server.
	 * @exception ObjectIOException when there is an I/O exception.
	 */
	public byte[] downloadInstrumentPackage(String instrument, float version)
		throws ObjectIOException, RemoteException {
		return stMgr_.downloadInstrumentPackage(instrument, version);
	}

	/**
	 * Check in the specified objects in a single atomic operation.
	 * 
	 * @param	usos	Descriptions of each StorableObject to be checked in.
	 * 
	 * @return		array of error strings.
	 *			If length = 0, no errors.
	 */
	public void checkIn(MultiObjSpec[] moi) throws ObjectIOException {
		stMgr_.checkIn(moi);
	}

	/**
	 * Check out the ObsBlk or CalBlk with the specified id plus
	 * any attached objects, ie Target, ObsDescr, etc.
	 * 
	 * @param	id	id of the ObsBlk or CalBlk to be checked out
	 * 
	 * @return		CheckOutStatus object that contains a return
	 *			status and (if successful) references to the
	 *			checked-out objects.
	 */
	public void checkOut(IdType[] idTypes)
		throws ObjectNotFoundException, ObjectIOException {
		stMgr_.checkOut(idTypes);
	}
    
    /**
     * Checkin the readme file for a given observing run
     * @param Readme file business object
     */
    
    public void checkInReadme(ObservingRun obsRun)
    throws ObjectIOException {
         stMgr_.checkInReadme(obsRun);

    }
    
    /**
     * Checkout the readme file for a given observing run.
     * @param boId Observing run definition. 
     */
    
    public String checkOutReadme(ObservingRun obsRun)
    throws ObjectIOException {
        return stMgr_.checkOutReadme(obsRun);
        
    }    
    
    
    
    

	/**
	 * Write several object to the external media in an atomic operation.
	 *      
	 * @param moi	Array of object descriptions.
	 * 
	 * @exception ObjectIOException
	 * 			if an I/O error occurred during the operation
	 * 
	 * @see #read
	 * @see #write
	 * @see #update
	 */
	public void multiStore(MultiObjSpec[] mos) throws ObjectIOException {
		stMgr_.multiStore(mos);
	}

	/**
	 * Returns a string representation of this object and its fields.
	 * 
	 * @return a string representation.
	 */
	public String toString() {
		return "RmiStorageManager["
			+ "auth="
			+ authenticated_
			+ ", mgr="
			+ stMgr_
			+ "]";
	}

    /* (non-Javadoc)
     * @see org.eso.ohs.phase2.apps.p2ppAppServer.RmiManagerInterface#usdSubmit(long)
     */
    public void usdSubmit(ObservingRun obsRun) throws ObjectIOException, RemoteException {
        stMgr_.usdSubmit(obsRun);
    }

    /* (non-Javadoc)
     * @see org.eso.ohs.phase2.apps.p2ppAppServer.RmiManagerInterface#populateReadmeFile(org.eso.ohs.dfs.ObservingRun)
     */
    public Readme populateReadmeFile(ObservingRun obsRun) throws ObjectIOException, RemoteException {
        return stMgr_.populateReadmeFile(obsRun);        
    }

}
