/* ===========================================================
 * JFreeChart : a free chart library for the Java(tm) platform
 * ===========================================================
 *
 * (C) Copyright 2000-2005, by Object Refinery Limited and Contributors.
 *
 * Project Info:  http://www.jfree.org/jfreechart/index.html
 *
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by 
 * the Free Software Foundation; either version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License 
 * along with this library; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc. 
 * in the United States and other countries.]
 *
 * ------------------
 * BarChartDemo1.java
 * ------------------
 * (C) Copyright 2003-2005, by Object Refinery Limited and Contributors.
 *
 * Original Author:  David Gilbert (for Object Refinery Limited);
 * Contributor(s):   ;
 *
 * $Id: TimelineView.java,v 1.4 2005/11/10 10:19:06 amchavan Exp $
 *
 * Changes
 * -------
 * 09-Mar-2005 : Version 1, copied from the demo collection that ships with
 *               the JFreeChart Developer Guide (DG);
 *
 */

package org.eso.ohs.phase2.timeline;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GradientPaint;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

import org.eso.ohs.core.utilities.Convert;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.CategoryAxis;
import org.jfree.chart.axis.DateAxis;
import org.jfree.chart.labels.IntervalCategoryToolTipGenerator;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.renderer.category.CategoryItemRenderer;
import org.jfree.chart.renderer.category.GanttRenderer;
import org.jfree.chart.urls.StandardCategoryURLGenerator;
import org.jfree.data.category.IntervalCategoryDataset;
import org.jfree.data.gantt.Task;
import org.jfree.data.gantt.TaskSeries;
import org.jfree.data.gantt.TaskSeriesCollection;
import org.jfree.data.time.SimpleTimePeriod;

/**
 * Widget to display a {@link Timeline}. It is a
 * {@link javax.swing.JPanel} (inherited via ChartPanel)
 * suitable for incorporating in a
 * GUI, for instance using the
 * {@link javax.swing.JFrame#setContentPane(java.awt.Container)}
 * method.
 * <p>
 * For instance:
 * <pre>
        // Prepare list of tasks, then compute a timeline
        TimelineTask[] tasks = ...
        Timeline timeline = new EarlyGreedy( tasks, startDate, endDate );
        
        // Setup a view for the timeline and display it in a
        // top-level window 
        TimelineView view = new TimelineView( timeline );
        JFrame      frame = new JFrame( "Example timeline" );
        frame.setContentPane( view );
        frame.pack();
        frame.setVisible( true );
 * </pre>
 * 
 * @version $Revision: 1.4 $
 * @author amchavan, Aug 31, 2005
 */

// $Id: TimelineView.java,v 1.4 2005/11/10 10:19:06 amchavan Exp $

public class TimelineView extends ChartPanel {

    /**
     * Create a Gantt-chart view of the input {@link Timeline}
     * 
     * @param timeline  
     */
    public TimelineView( Timeline timeline ) {
        super( null, false );
        IntervalCategoryDataset dataset = createDataset( timeline );
        JFreeChart chart = createGanttChart( dataset );
        setChart( chart );
        setPreferredSize( new Dimension( 500, 270 ) );
    }

    /**
     * @return The dataset for the Gantt chart, created by converting all
     *         TimelineTasks in the input Timeline to the chart's own
     *         tasks.
     */
    private static IntervalCategoryDataset createDataset( Timeline timeline ) {

        TaskSeries s1 = new TaskSeries( "Scheduled" );

        // Map all our TimelineTasks to the chart's Tasks
        //-----------------------------------------------
        
        TimelineTask[] tasks = timeline.getSchedTasks();
        for( int i = 0; i < tasks.length; i++ ) {
            
            TimelineTask ttask = tasks[i];
            
            // Compute the new Task's description
            StringBuffer sb = new StringBuffer();
            sb.append( ttask.getID() )
              .append( ": '" )
              .append( ttask.getName() )
              .append( "'" );
            String desc = sb.toString();
            
            // Compute the new Task's duration
            Date tstart = ttask.getStartTime();
            Date   tend = new Date( ttask.getStartTime().getTime() +
                                    ttask.getDuration() * 1000 );
            SimpleTimePeriod period = new SimpleTimePeriod( tstart, tend );
            
            // Create the new Task and add it to the dataset
            Task gtask = new Task( desc, period );
            s1.add( gtask );
        }

        TaskSeriesCollection collection = new TaskSeriesCollection();
        collection.add( s1 );

        return collection;
    }

    /**
     * Create a Gantt chart.
     * 
     * @return The chart.
     */
    private static JFreeChart createGanttChart( IntervalCategoryDataset dataset ) {
        JFreeChart chart = createGanttChart( 
                null,   // "Gantt Chart",
                null,   // "Domain",
                null,   // "Range",
                dataset, 
                false,  // include legend
                true,   // do tooltips
                true ); // do generate URLs

        chart.setBackgroundPaint( Color.white );
        
         // get a reference to the plot for further customisation...
        CategoryPlot plot = chart.getCategoryPlot();
        plot.setBackgroundPaint( Color.LIGHT_GRAY );
        plot.setDomainGridlinePaint( Color.white );
        plot.setDomainGridlinesVisible( true );
        plot.setRangeGridlinePaint( Color.white );

        // set up gradient paints for series...
        Color start = Color.blue;
        Color   end = Color.green;//new Color( 0, 0, 64 );
        GradientPaint gp0 = new GradientPaint(    0.0f, 0.0f, start, 
                                               1000.0f, 0.0f, end, 
                                               false );
         GanttRenderer renderer = (GanttRenderer) plot.getRenderer();
         renderer.setSeriesPaint(0, gp0);
        return chart;
    }
    
    /**
     * [Originally taken from JFreeChart's  
     * <code>ChartFactory.createGanttChart()</code>]
     * <p>
     * Creates a Gantt chart using the supplied attributes plus default values 
     * where required.  The chart object returned by this method uses a 
     * CategoryPlot instance as the plot, with a CategoryAxis 
     * for the domain axis, a DateAxis as the range axis, and a 
     * GanttRenderer as the renderer.
     *
     * @param title  the chart title (<code>null</code> permitted).
     * @param categoryAxisLabel  the label for the category axis 
     *                           (<code>null</code> permitted).
     * @param dateAxisLabel  the label for the date axis 
     *                       (<code>null</code> permitted).
     * @param dataset  the dataset for the chart (<code>null</code> permitted).
     * @param legend  a flag specifying whether or not a legend is required.
     * @param tooltips  configure chart to generate tool tips?
     * @param urls  configure chart to generate URLs?
     *
     * @return A Gantt chart.
     */
    private static JFreeChart createGanttChart( String title,
                                                String categoryAxisLabel,
                                                String dateAxisLabel,
                                                IntervalCategoryDataset dataset,
                                                boolean legend,
                                                boolean tooltips,
                                                boolean urls) {

        CategoryAxis categoryAxis = new CategoryAxis( categoryAxisLabel );
        DateAxis dateAxis = new DateAxis( dateAxisLabel );

        CategoryItemRenderer renderer = new GanttRenderer();
        if (tooltips) {
            DateFormat isofmt = new SimpleDateFormat( Convert.ISO_DATE_FORMAT );
            renderer.setBaseToolTipGenerator(
                new IntervalCategoryToolTipGenerator( "{3}", isofmt ));
        }
        if (urls) {
            renderer.setBaseItemURLGenerator( 
                    new StandardCategoryURLGenerator() );
        }

        CategoryPlot plot = new CategoryPlot(
            dataset, categoryAxis, dateAxis, renderer );
        plot.setOrientation(PlotOrientation.HORIZONTAL);
        JFreeChart chart = new JFreeChart(
            title, JFreeChart.DEFAULT_TITLE_FONT, plot, legend );

        return chart;
    }

    /**
     * Makes this view display a new timeline; the old one is discarded
     * and lost.
     * @param timeline  The new Timeline to display
     */
    // amchavan, 10-Nov-2005
    public void setTimeline( Timeline timeline ) {
        JFreeChart chart = getChart();
        CategoryPlot plot = (CategoryPlot) chart.getPlot();
        IntervalCategoryDataset dataset = createDataset( timeline );
        plot.setDataset(dataset);
    }
}
