package org.eso.ohs.phase2.visibility;

/*
 * 
 * This class computes visibilities of observation blocks
 * 
 * Created on Jun 7, 2005
 * @author pnunes and ddorigo
 *
 */

import java.util.Date;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.Convert;
import org.eso.ohs.dfs.TimeInterval;
import org.eso.ohs.phase2.orang.OBScheduleInfo;
import org.eso.ohs.phase2.orang.ObDataContainer;
import org.eso.ohs.phase2.visibility.PCFPlotterFilter.ValueRejectedException;
import org.eso.ohs.phase2.visiplot.TimerLog;

import com.mhuss.AstroLib.AstroOps;
import com.mhuss.AstroLib.Latitude;
import com.mhuss.AstroLib.Longitude;
import com.mhuss.AstroLib.NoInitException;
import com.mhuss.AstroLib.ObsInfo;
import com.mhuss.AstroLib.PlanetData;
import com.mhuss.AstroLib.Planets;



public class VisibilityJCalc implements VisibilityCalc {
	 
	/**
	 * Visibility data cache 
	 */
	VisibilityCache visibilityCache_ = VisibilityCache.getInstance();
	
	
	/**
     * Logging instance.
     */
    protected static Logger stdlog_ =
        Logger.getLogger(VisibilityJCalc.class);

	/*
	 * Calculates the visibility information for an array of OBs for a period of time	 *
	 *   	 
	 * @param obs  OB data container objects
	 * @param interval  (minutes) 
	 * @param startTime
	 * @param endTime
	 * @param observLatitude ( Latitude of the observatory )
	 * @param observLongitude ( Longitude of the observatory ) 
	 * @return array of OBScheduleInfo objects
	 */
	public OBScheduleInfo [] calc(ObDataContainer [] obs , int interval,
			Date startTime, Date endTime, 
			VisicalcOptions options, double observLatitude, double observLongitude ) throws ValueRejectedException  {
	    String[] splitString = {"sun/moon","sun/moonPCF","airmass, moon dist","time intervals","sidereal Time Intervals", "PCFs","Confine","zenithConstraint","OBInfo"};
		TimerLog timer = new TimerLog(splitString);
		timer.setCalculate(true);
		
		
		double jDate = 0.0;
		double targetRA = 0.0;
		double targetDEC = 0.0;
		double sunAngle = 0.0;
		double fli = 0.0;
//		double moonRA = 0.0;
//		double moonDec = 0.0;
		double moonDist = 0.0;
		double moonEl = 0.0; 
		double moonUp = 0.0 ; // 1.0 if up
		double airMass = 0.0;
		double timeInterval = 0.0;
		double lst = 0.0;
		
		
		double moonDistanceTolerance = options.getMoonDistanceTolerance(); 
		double airmassTolerance = options.getAirmassTolerance();
		double fliTolerance = options.getFliTolerance();
		double sunConstraint = options.getSunConstraint();
		double moonConstraint = options.getMoonConstraint();
		
		boolean calcFli = options.calcFli();
		boolean calcMoonDist = options.calcMoonDist();
		boolean calcMoon = options.calcMoon();
		boolean calcSun = options.calcSun();
		boolean calcTimeInterval = options.calcTimeInterval();
		boolean calcSiderealTimeInterval = options.calcSiderealTimeInterval();
		boolean calcZenithConstraint = options.calcZenithInvisibility();
		
		double [] moonElevations;
		double [] moonUps;
		double [] sunAngles;
		double [] flis;
		double [] timeIntervals;
		double [] siderealTimeIntervals;
		double [] airMasses;
		double [] moonDistances;
		double [] meetZenithConstraint;
		double [] execution;
		double [] result ;
		double [] lsts;
		
		
		OBScheduleInfo [] resInfo = new OBScheduleInfo[obs.length];
		
		// sets the upper limmit of iteration
		int total = (60 / interval) * VisibilityCalcUtils.getDuration(startTime, endTime);
		
		if (total <= 0) {
			total = (60 / interval) * VisibilityCalcUtils.DURATION;
		}
				
		
		moonElevations = new double[total];		
		sunAngles = new double[total];
		moonUps = new double[total];			
		flis = new double[total];
		meetZenithConstraint = new double[total];
		lsts = new double[total];

		
		
		//Iterate through the interval and
		//calculate Moon positions, sun and moon angle 		 	
		double [] julianDates = new double[total];
		
		for (int i = 0; i < total; i++) {
			timer.calcTimer(0);
			
			// Calculate the Julian Date plus the interval time
			jDate = VisibilityCalcUtils.getJulianDate(startTime, interval, i);
			julianDates[i] = jDate;
			
			// local sidereal time
			lst = getLST(jDate,observLongitude);
			lsts[i] = lst;

			// 	sun elevation
			if ( calcSun) {
				sunAngle = getSunValue(jDate,observLongitude,observLatitude,lst);				
			}
			
			// moon calculations
			if (calcMoon) {
				
				// moon positions (assuming caching)
				
				// moon elevation
				moonEl = getMoonAngle(jDate, observLongitude, observLatitude);
				moonUp = getMoonVisible(jDate,observLongitude,observLatitude);
				
				if (calcFli) {		
					fli = getFLIValue(jDate,observLongitude,observLatitude);										
				}

			}
			
			sunAngles[i] = sunAngle;
			moonElevations[i] = moonEl;
			moonUps[i] = moonUp;
			flis[i] = fli;	
			timer.calcTimer(0);
	
		}
		
		timer.calcTimer(1);
		
		
		//	 create the PCFs from the arrays		
		PCF moonUpPCF = new PCF();		
		if ( calcMoon ) {						       			
			moonUpPCF.buildFromArrayConstraint(startTime,interval,moonElevations,moonConstraint,true);			
		}
		
		PCF sunPCF = new PCF() ;
		if ( calcSun) {			
			sunPCF.buildFromArrayConstraint(startTime,interval,sunAngles,sunConstraint,false);
		}
		timer.calcTimer(1);
				
		// Calculate airmass and moon angle for each Observation Block
		// and store this information on the OBScheduleInfo
		for (int i = 0; i < obs.length; i++) {
            
            stdlog_.debug( " Calculate visibility constraint "+i);
         
		    ObDataContainer ob = obs[i];
			
			/// converts
			targetRA = ob.getObInfo().getRa()* VisibilityCalcUtils.MILLARC_HRMMSS;
	        targetDEC = (ob.getObInfo().getDec() / VisibilityCalcUtils.MILLARC_MMSS) / VisibilityCalcUtils.MILLISECS_PER_SEC;

			airMasses = new double[julianDates.length];
			moonDistances = new double[julianDates.length];
			timeIntervals = new double[julianDates.length];
			siderealTimeIntervals = new double[julianDates.length];

		
			
			// for each julian date
			for (int j = 0; j < julianDates.length; j++) {
				timer.calcTimer(2);
				
				jDate = julianDates[j];				
				
				// calculate local sidereal time
				lst =  getLST(jDate,observLongitude);
				
				// calculate airmass
				airMass = getAirMassValue(jDate,lst,targetRA,targetDEC,observLatitude);
				airMass = Convert.round(airMass, 6);								
				airMasses[j]= airMass;
							
								
					
				// calculate moon angular distance with target
				if ( calcMoonDist ) {				
					moonDist = getMoonDistValue(jDate,targetRA,targetDEC,observLongitude,observLatitude);
					moonDistances[j] = moonDist;
				}	
				
				Date utDate = new Date(startTime.getTime() + (interval * j * 60000));
				
				// calculate if the date falls into user defined time intervals
				timer.calcTimer(2);
				timer.calcTimer(3);
				if ( calcTimeInterval) {
					timeInterval = VisibilityCalcUtils.getTimeIntervalValue(ob.getTimeIntervals(), utDate);				
					timeIntervals[j] = timeInterval;
				}
				timer.calcTimer(3);
				timer.calcTimer(4);
				
				// calculate if the date falls into user defined sidereal time intervals
				if ( calcSiderealTimeInterval ) {
					timeInterval = getSiderealTimeIntervalValue(ob.getSiderealTimeIntervals(), utDate, jDate);
					siderealTimeIntervals[j] = timeInterval;
				}
				timer.calcTimer(4);				
				
			}
			
			timer.calcTimer(5);
			
			// 	build the PCFs
			
			
			// time interval
			PCF timeIntervalPCF = new PCF();
			
			if (calcTimeInterval) {				
				timeIntervalPCF.buildFromArray(startTime,interval,timeIntervals);
			}
			
			
			//	sidereal time interval
			PCF siderealTimeIntervalPCF = new PCF();
			
			if ( calcSiderealTimeInterval ) {				
				siderealTimeIntervalPCF.buildFromArray(startTime,interval,siderealTimeIntervals);
			}

			
			// moon pcfs
			PCF moonDistancePCF = new PCF();
			PCF fliPCF = new PCF();
			double moonDistanceConstraint =   ob.getObInfo().getMoonAngularDistance() * (( 100 - moonDistanceTolerance  ) / 100);
			double fliConstraint =   ob.getObInfo().getFractionalLunarIllumination() * (( 100 + fliTolerance  ) / 100);
			
			if (calcMoon) {
						
				// moon distance
				if ( options.calcMoonDist() ) {
					 
					double [] moonDistanceBinary = getMoonDistanceBinary(moonDistances,moonDistanceConstraint,moonUps);
					moonDistancePCF.buildFromArray(startTime,interval,moonDistanceBinary);
				}				
				// fli				
				if ( options.calcFli() ) {	
					fliPCF.buildFromArrayConstraint(startTime,interval,flis,fliConstraint,false);				
				}
			}
			
			//zenith constraint
			PCF meetsZenithConstraintPCF = new PCF();
			meetsZenithConstraintPCF.buildFromArray(startTime,interval,meetZenithConstraint);
			
			// airmass
			double airmassConstraint = ob.getObInfo().getAirmass() * (( 100 + airmassTolerance  ) / 100);
			
	        PCF airmassPCF = new PCF();	        
	        airmassPCF.buildFromArray(startTime,interval,airMasses);

			
	        timer.calcTimer(5);
	        
	        timer.calcTimer(6);
	        
	        
	        // result pcf 
	             
			
	        result = getResult(startTime, interval, airMasses,
					moonDistances, sunAngles, timeIntervals,
					siderealTimeIntervals, flis, moonUps,meetZenithConstraint,
					airmassConstraint,moonDistanceConstraint,fliConstraint, 
					options);	
	        
	        timer.calcTimer(6);
	        
	        timer.calcTimer(7);
	        
	        // apply zenith constraint
	        
	        if (calcZenithConstraint) {
	        	
	        	if ( entersZenithRegion(options.getZenithConstraint(),targetDEC, observLatitude) ) {
	        		double [] startEndZenith = findZenithAreaAnalitical(options.getZenithConstraint(),targetRA,targetDEC,observLatitude,observLongitude);
	        		result = confineZenithConstraint(lsts,startEndZenith[0],startEndZenith[1],result);
	        	}
	        }
	        
	        
	        timer.calcTimer(7);
	        
	        timer.calcTimer(8);
	        
	        PCF resultPCF = new PCF();
		    resultPCF.buildFromArray(startTime,interval,result);		    
	        
	        OBScheduleInfo info = new OBScheduleInfo();
	        execution = calcExecTimes(info,result,interval,startTime, ob.getObInfo().getExecutionTime());
	        
	        PCF executionPCF = new PCF();
	        executionPCF.buildFromArray(startTime,interval,execution);
	        
			// add information to info						
			info.setOb(ob);								
			info.setAirmassPCF(airmassPCF);
			info.setMeetsZenithConstraintPCF(meetsZenithConstraintPCF);
			info.setMoonAnglePCF(moonDistancePCF);
			info.setFliPCF(fliPCF);
			info.setMoonUpPCF(moonUpPCF);
			info.setSunDownPCF(sunPCF);			
			info.setResultPCF(resultPCF);									
			info.setTimeIntervalPCF(timeIntervalPCF);
			info.setSiderealTimeIntervalPCF(siderealTimeIntervalPCF);
			info.setExecTimePCF(executionPCF);
			info.setStart(startTime);
			info.setEnd(endTime);
			info.setInterval(interval);
			
			
			resInfo[i] = info;
			//System.out.println(info);
			timer.calcTimer(8);
		}
		
		stdlog_.debug("Visibility computation performance report:");
		stdlog_.debug(timer.reportResult());
				
		return(resInfo);
		
		
	}
	
	
	
	
	/**
	 * Returns a binary array that considers if the moonDistanceConstraint is fulfilled
	 * 
	 * @param moonDistances
	 * @param moonDistanceConstraint
	 * @param moonUps
	 * @return
	 */
	private double[] getMoonDistanceBinary(double[] moonDistances, double moonDistanceConstraint, double[] moonUps) {
		double [] res = new double[moonDistances.length];
		
		for (int i = 0; i < moonUps.length; i++) {				
			double up = moonUps[i];	    		
			if ( up == 1.0 ) 	
				if( moonDistances[i] <  moonDistanceConstraint ) 
					res[i] = 0;
				else 
					res[i] = 1;
			else 
				res[i] = 1;
		}
		
		return(res);
	}





	/**
	 * Determines if a target enters the zenith avoidance region
	 * 
	 * @param zenithConstraint in degrees
	 * @param targetDEC in radians
	 * @param observLatitude in radians
	 * @return
	 */
	private boolean entersZenithRegion(double zenithConstraint, double targetDEC, double observLatitude) {
		return(Math.abs(targetDEC - observLatitude)<=Math.toRadians(zenithConstraint));
	}



	
	/**
	 * 
	 * @param zenithConstraint
	 * @param targetRA
	 * @param targetDEC
	 * @param observLatitude
	 * @param observLongitude
	 * @return a double[2] array 
	 */
	protected double [] findZenithAreaAnalitical(double zenithConstraint, double targetRA, double targetDEC,
			double observLatitude, double observLongitude) {
		
		double [] res = new double[2];
				
		double ha = Math.abs(Math.acos( ( Math.cos(Math.toRadians(zenithConstraint))
				  - (Math.sin(observLatitude) * Math.sin(targetDEC)))
				/ (Math.cos(observLatitude) * Math.cos(targetDEC))   ));
		
		double lst1 = targetRA - ha;
		double lst2 = targetRA + ha;				
		
		res[0] = lst1;
		res[1] = lst2;
		
		return(res);
	}
	
	
 
	
	
	

	/**
	 * Returns the airmass for an object in (ra,dec) position, with relation to observatory
	 * latitude and local sidereal time
	 * 
	 * @param ra Right ascension (radians)
	 * @param dec Declination (radians)
	 * @param obsLat observatory latitude (radians)
	 * @param lst Local sidereal time
	 * @return Airmass
	 */
	protected double airmass(double ra, double dec, double obsLat, double lst) {
		
		// zenith angle
		double za = zenithAngle(ra, dec, obsLat, lst);
		
		
		// airmass
		double airmass;
		
		/*
		 * Copied from SLALIB fortran source to calculate the airmass based on zenith angle
		 * 
		 * SECZM1 = 1D0/(COS(MIN(1.52D0,ABS(ZD))))-1D0 sla_AIRMAS = 1D0 +
		 * SECZM1*(0.9981833D0 : - SECZM1*(0.002875D0 + 0.0008083D0*SECZM1))
		 */
		
		double secz = 1 / (Math.cos(Math.min(1.52, Math.abs(za)))) - 1;
		airmass = 1 + secz * (0.9981833 - secz * (0.002875 + 0.0008083 * secz));
		
		return (airmass);
	}
	
	/**
	 * Calculates the zenith angle for an object in (ra,dec) position, 
	 * with relation to observatory latitude and local sidereal time
	 * 
	 * @param ra  
	 * @param dec  (radians)
	 * @param obsLat (radians)
	 * @param lst
	 * @return double (radians)
	 */
	protected double zenithAngle(double ra, double dec, double obsLat,
			double lst) {
			
		
		double el = getElevation(ra, dec, obsLat, lst);

		// azimuth angle (radians)
		//double az = Math.atan2(-Math.sin(ha) * Math.cos(dec), Math.cos(obsLat)
		//		* Math.sin(dec) - Math.sin(obsLat) * Math.cos(dec)
		//		* Math.cos(ha));

		
		// zenith angle
		double za = (Math.PI / 2) - el;
		return za;
	}


	/**
	 *
	 * Returns the target's elevation, given declination, latitude and hour angle
	 * 
	 * @param targetRA
	 * @param targetDec
	 * @param obsLat
	 * @param lst
	 * @return
	 */
	protected double getElevation(double targetRA, double targetDec, double obsLat, double lst) {
		//		 hour angle
		double ha = lst - targetRA;	
		
		// elevation (radians)
		double el = Math.asin(Math.sin(obsLat) * Math.sin(targetDec)
				+ Math.cos(obsLat) * Math.cos(targetDec) * Math.cos(ha));
		return el;
	}
	
	
	
	/* (non-Javadoc)
	 * @see org.eso.ohs.phase2.visiplot.VisibilityCalc#getAirMassValue(double, double, double, double, double)
	 */
	public double getAirMassValue(double jdate, double lst, double targetRA,
			double targetDEC, double lat) {
		
		double res = airmass(targetRA, targetDEC, lat, lst);
		return res;
	}
	
	
	
	/* (non-Javadoc)
	 * @see org.eso.ohs.phase2.visiplot.VisibilityCalc#getFLIValue(double, double, double)
	 */
	public double getFLIValue(double jdate, double longit, double lat) {
		
		
		double d = 0;
		double fli = 0;
		
		Latitude latitude = new Latitude(Math.toDegrees(lat));
		Longitude longitude = new Longitude(Math.toDegrees(-longit));			
		
		try {		
			ObsInfo oi = new ObsInfo(latitude, longitude, 0);
			
			// obtain RA and DEC for the moon
			PlanetData pd = new PlanetData(Planets.LUNA, jdate, oi);
			
			double moonRA = AstroOps.normalizeRadians(pd.getRightAscension());
			double moonDec= pd.getDeclination();
			
			
			//	obtain RA and DEC for the sun
			pd = new PlanetData(Planets.SUN, jdate, oi);
			double sunRA = AstroOps.normalizeRadians(pd.getRightAscension());
			double sunDec= pd.getDeclination();
			
			
			// calculate angular distance between the two objects
			d = angularDistance(moonRA, moonDec, sunRA, sunDec);
			
			// calculate FLI based on this distance
			fli = (1.0 - Math.cos(d))/2;
			
		} catch (NoInitException e) {
			e.printStackTrace();
		}
		
		return fli;
	}
	
	
	/**
	 * Calculates the angular distance between two objects
	 * Took from Duffet-Smith book
	 * 
	 * @param ra1  RA of the first object
	 * @param dec1 DEC of the first object
	 * @param ra2 RA of the second object
	 * @param dec2 DEC of the secon object
	 * @return Angular distance
	 */
	private double angularDistance(double ra1, double dec1, double ra2, double dec2) {
		double d;
		
		// calculate the angular distance
		double cosD = Math.sin(dec1)*Math.sin(dec2) + Math.cos(dec1)*Math.cos(dec2)*Math.cos(ra1-ra2);
		
		d = Math.acos(cosD);
		return d;
	}
	
	/*
	 * (non-Javadoc)
	 * @see org.eso.ohs.phase2.visiplot.VisibilityCalc#getLST(double, double)
	 */
	public double getLST(double jdate, double longitude) {
		
		Double lst;		
		
		lst = visibilityCache_.getLST(jdate);		
		
		if (lst == null ) {
			
			double gmst = AstroOps.greenwichSiderealTime(jdate);
			gmst = AstroOps.normalizeRadians(gmst);
			lst = new Double(gmst - longitude);		
			
			// store it in cache
			visibilityCache_.putLST(jdate,lst.doubleValue());
			
		}
		return (lst.doubleValue());
		
	}
	
	
	
	/* (non-Javadoc)
	 * @see org.eso.ohs.phase2.visiplot.VisibilityCalc#getMoonDistValue(double, double, double, double, double)
	 */
	public double getMoonDistValue(double jdate, double targetRA,
			double targetDEC, double longit, double lat) {
		
		double distance = 0;
		double moonRA;
		double moonDec;
		
		moonRA = getMoonRa(jdate,longit,lat);		
		moonDec = getMoonDec(jdate,longit,lat);		
		
		
		moonRA =  Math.toRadians(moonRA);
		moonDec = Math.toRadians(moonDec);
		
		
		distance = angularDistance(moonRA,moonDec,targetRA,targetDEC);
		
		distance = Math.toDegrees(distance); 
		return distance;
	}
	
	
	
	
	
	/* (non-Javadoc)
	 * @see org.eso.ohs.phase2.visiplot.VisibilityCalc#getSunValue(double, double, double, double)
	 */
	public double getSunValue(double jdate, double longit, double lat,
			double lst) {
		
		Double sunAngle = null;
		double tmp = 0;
		
		//	Check the cache			
		sunAngle = visibilityCache_.getSunElevation(jdate);
		
		// check if it is cached
		if (sunAngle == null) {
			
			// calculate 
			Latitude latitude = new Latitude(Math.toDegrees(lat));
			Longitude longitude = new Longitude(Math.toDegrees(-longit));
			ObsInfo oi = new ObsInfo(latitude, longitude);
			PlanetData pd = new PlanetData(Planets.SUN, jdate, oi);		
			
			try {
				tmp = Math.asin(pd.getAltAzRadius()); //- risesetAlt;;
				
				//sunAngle = AstroOps.normalizeRadians(sunAngle);
			} catch (NoInitException e) {
				e.printStackTrace();
			}
			
			sunAngle  = new Double(Math.toDegrees(tmp));
			
			
			// store it in cache
			visibilityCache_.putSunElevation(jdate,sunAngle.doubleValue());		
		}
		
		return(sunAngle.doubleValue());
	}
	
	
	/*
	 * (non-Javadoc)	 
	 * @see org.eso.ohs.phase2.visiplot.VisibilityCalc#getMoonVisible(double,
	 *      double, double)
	 */
	public double getMoonVisible(double jdate, double longit, double lat) {
		
		double moonAngle = getMoonAngle(jdate, longit, lat);
		boolean visible = moonAngle >= 0.0;
		
		if (visible)
			return (1);
		else
			return (0);
	}
	
	
	/* (non-Javadoc)
	 * @see org.eso.ohs.phase2.visiplot.VisibilityCalc#getMoonAngle(double, double, double)
	 */
	public double getMoonAngle(double jdate, double longit, double lat) {
		
		Double moonAngle = null;
		double tmp = 0;
		
		//	 Check the cache
		moonAngle = visibilityCache_.getMoonElevation(jdate);
		
		// check if it is cached
		if (moonAngle == null) {
			
			// calculate moon angle
			Latitude latitude = new Latitude(Math.toDegrees(lat));
			Longitude longitude = new Longitude(Math.toDegrees(-longit));
			ObsInfo oi = new ObsInfo(latitude, longitude, 0);
			PlanetData pd = new PlanetData(Planets.LUNA, jdate, oi);									
			try {
				tmp = Math.asin(pd.getAltAzRadius());			
			} catch (NoInitException e) {
				e.printStackTrace();
			}
			moonAngle =  new Double(Math.toDegrees(tmp));
			
			// store it in cache
			visibilityCache_.putMoonElevation(jdate,moonAngle.doubleValue());
			
		}
		
		return(moonAngle.doubleValue());
	}
	
	/* (non-Javadoc)
	 * @see org.eso.ohs.phase2.visiplot.VisibilityCalc#getMoonAngle(double, double, double)
	 */
	public double getMoonRa(double jdate, double longit, double lat) {
		
		double tmp = 0;
		Double moonRa = null;
		
		// Check the cache
		moonRa = visibilityCache_.getMoonRa(jdate);
		
		if (moonRa == null) {
			
			// calculate it
			Latitude latitude = new Latitude(Math.toDegrees(lat));
			Longitude longitude = new Longitude(Math.toDegrees(-longit));
			
			ObsInfo oi = new ObsInfo(latitude, longitude, 0);
			PlanetData pd = new PlanetData(Planets.LUNA, jdate, oi);
			
			
			try {
				tmp = pd.getRightAscension();
			} catch (NoInitException e) {
				e.printStackTrace();
			}
			
			tmp = Math.toDegrees(tmp);
			tmp = AstroOps.normalizeDegrees(tmp);
			
			moonRa = new Double(tmp);
			
			// store it in cache
			visibilityCache_.putMoonRa(jdate,moonRa.doubleValue());
		}
		
		return moonRa.doubleValue();
	}
	
	/* (non-Javadoc)
	 * @see org.eso.ohs.phase2.visiplot.VisibilityCalc#getMoonAngle(double, double, double)
	 */
	public double getMoonDec(double jdate, double longit, double lat) {
		
		double tmp = 0;
		Double moonDec = null;
		
		// Check the cache
		moonDec = visibilityCache_.getMoonDec(jdate);
		
		if (moonDec == null ) {
			
			Latitude latitude = new Latitude(Math.toDegrees(lat));
			Longitude longitude = new Longitude(Math.toDegrees(-longit));
			
			ObsInfo oi = new ObsInfo(latitude, longitude, 0);
			PlanetData pd = new PlanetData(Planets.LUNA, jdate, oi);
			
			try {
				tmp = pd.getDeclination();
			} catch (NoInitException e) {
				e.printStackTrace();
			}
			tmp = Math.toDegrees(tmp);			
			moonDec = new Double(tmp);
			
			// store in cache
			visibilityCache_.putMoonDec(jdate,moonDec.doubleValue());
			
		}
		
		return(moonDec.doubleValue());
	}
	

	
	
	 /**
     * getTimeIntervalValue - get the Sidereal Time Interval Value If the time
     * interval has not been defined in the OB any time period is satisfactory
     * set the returned value to 1. Else if the start and end time does not fall
     * within this interval time (precision) set returned value to 0.
     * @param ObservationBlock
     * @param julian Date double
     * @return time interval value
     */
    public double getSiderealTimeIntervalValue(
    		TimeInterval[] timeInts,
        Date utDate,
        double jdate) {

        // Convet the jdate to LST in radioans.

        //Convert user values into radian
        //Start and end dates for ST Times.
        // If the end date is < start date then 
        // add 2 PI to end date. 

        // Then compate the values         
        double timeValue = 1.;
      
        
        double siderealTimeInRadians = this.getLST(jdate,VisibilityCalcUtils.PLONG);
        if (siderealTimeInRadians < 0) {
            siderealTimeInRadians = siderealTimeInRadians + (2 * Math.PI);
        }
        if (siderealTimeInRadians > (2 * Math.PI)) {
            siderealTimeInRadians = siderealTimeInRadians - (2 * Math.PI);
        }
        if (timeInts != null && timeInts.length != 0) {
            for (int i = 0; i < timeInts.length; i++) {
                double startTime =
                    timeInts[i].getStartTime() * (Math.PI / (12 * 3600));
                double endTime = timeInts[i].getEndTime() * (Math.PI / (12 * 3600));
                if (((siderealTimeInRadians >= startTime)
                    && (siderealTimeInRadians <= endTime))
                       || ((siderealTimeInRadians + (2 * Math.PI) >= startTime)
                        && (siderealTimeInRadians + (2 * Math.PI) <= endTime))) {
                    timeValue = 1.;
                    return (timeValue);
                } else {
                    timeValue = 0.;

                }
            }
        }
        return (timeValue);
    }
    
    /**
     * Iterates through the airmass array, 
     * and determines the intervals within which an OB is observable.
     * Adds these results to schedInfo object, 
     * and cuts the non-observable periods from the result
     * It returns an array with the execution time periods
     * 
     * @param schedInfo
     * @param resultPCF
     * @param execTime of the OB in seconds
     * @return 
     */
    protected double [] calcExecTimes(OBScheduleInfo schedInfo , double [] result, int interval , Date startDate,  int execTime) {  	
        int msecsInterval = VisibilityCalcUtils.MILLISECS_PER_SEC * 60 * interval;    	                       
        int length = result.length;                
        int execTimeMilli = execTime* VisibilityCalcUtils.MILLISECS_PER_SEC;                
        boolean inside = false; // whether if you are inside a visible interval or not
        
        double [] execTimeRes = new double[length]; 
        
        Date start = null;
        Date end = null;
        
        Date currDate = new Date(startDate.getTime());
        
        int i = 0;
        //find intervals 
        for (i = 0; i < length; i++) {
            double value = result[i];            
            execTimeRes[i] =  0;
            
            // found a visible point
            if (value >= 0  && value < AirMass.NOT_VIS) {
            	if (!inside ) {
            		start = currDate;            		
            		inside = true;            		
            	}

            // not visible 
            } else {            	             	
            	// found the end of a visible interval
            	if ( inside ) {
            		inside = false;	     	
            		end = currDate;
            		
            		markNotVisible(interval, i, execTime, execTimeRes) ;
            		
            		
            		// check if the execution time fits the interval
            		long intervalSize = end.getTime() - start.getTime();            		            		
            		if ( intervalSize >= execTimeMilli ) {
            			
            			// it fits, lets define when can the OB start being executed within the interval            			
            			end = new Date( end.getTime() - execTimeMilli  );
            			schedInfo.addObservableInterval(start,end);            			
            			            			
            			
            			
            		}
            		
            		

            	}		            	
            }

            currDate = new Date(currDate.getTime() + msecsInterval  );
        }
        
        // end of the loop, if its inside a visible interval, insert it 
        if ( inside ) {
        	end = currDate;
        	
        	markNotVisible(interval, i, execTime, execTimeRes) ;
        	
        	// 	 check if the execution time fits the interval
    		long intervalSize = end.getTime() - start.getTime();            		            		
    		if ( intervalSize >= execTimeMilli ) {
    			//  it fits, lets define when can the OB start being executed within the interval            			
    			end = new Date( end.getTime() - execTimeMilli );
    			schedInfo.addObservableInterval(start,end);
    			
    			
    			
    		}
        }
        
        
        return execTimeRes;
        
    }
    

    /**
     * Changes execution array to mark a part of it not visible, given a step interval,
     * execution time needed and an index to start from
     * It marks not visibile from that index back  
     * 
     * @param interval in minutes
     * @param index 
     * @param execTime in seconds
     * @param execTimeRes array to be 
     */
    private void markNotVisible(int interval, int index, int execTime, double [] execTimeRes) {
 
    	int execTimeMins = execTime/60;
    	
		int grayAreaIndex = (int)(index -  execTimeMins/interval );
		if (grayAreaIndex >= 0) {
			for (int j = grayAreaIndex; j <= index; j++) {
				if ( j < execTimeRes.length )
					execTimeRes[j] = 1;
			}
		}
    }
    
    /**
	 * Combine all of the selected PCFs
	 * @return result pcf
	 */
	private double [] getResult( Date start, int interval,
			double [] airmass,
			double [] moonDist,
			double [] sun,
			double [] timeInterval,
			double [] siderealTimeInterval,
			double [] fli,
			double [] moonUp,
			double [] meetsZenithConstraint,
			double airmassOBConstraint,
			double moonOBDistConstraint,
			double fliOBConstraint,
			VisicalcOptions options
		) {
		
		// initialize airmass array
	    double [] res = new double[airmass.length];
	    for (int i = 0; i < res.length; i++) {
	    	res[i] = 1;
		}
	
	    // if the moon is taken into account
	    if (options.calcMoon()) {
	        	     
	    
	    	// create binarys for moon distance and FLI
	    	double [] moonDistBinary = new double[moonDist.length];
	    	double [] fliBinary = new double[fli.length];	    
	    	
	    	for (int i = 0; i < fli.length; i++) {			
	    		if ( fli[i] > fliOBConstraint )
	    			fliBinary[i] = 0;
	    		else
	    			fliBinary[i] = 1;			
	    		if ( moonDist[i] < moonOBDistConstraint )
	    			moonDistBinary[i] = 0;
	    		else
	    			moonDistBinary[i] = 1;	
	    	}
	    	
	    	
	    	// when the moon is down the value should be untouched, otherwise is the
	    	// conjunction of both FLI and moonDistance	    	
	    	for (int i = 0; i < moonUp.length; i++) {				
	    		double up = moonUp[i];	    		
	    		if ( up == 1.0 ) {	    		
	    			if (options.calcFli()) {
	    				res[i] = res[i] * fliBinary[i];
	    			}		        
	    			if (options.calcMoonDist()) {
	    				
	    				res[i] = res[i] * moonDistBinary[i];
	    			}
	    		}
	    	}
	    }
	    
	    //creates the sun binary array
	    double [] sunDownBinary = new double[sun.length];
	    for (int i = 0; i < sunDownBinary.length; i++) {
			
	    	if ( sun[i] < options.getSunConstraint())
	    		sunDownBinary[i] = 1;
	    	else
	    		sunDownBinary[i] = 0;
		}
	    
	    if (options.calcSun()) {
        	res = VisibilityCalcUtils.conjunction(res,sunDownBinary);
        }
	    
	    if (options.calcTimeInterval()) {
	    	res = VisibilityCalcUtils.conjunction(res,timeInterval);
	    }
	    if (options.calcSiderealTimeInterval()) {
	    	res = VisibilityCalcUtils.conjunction(res,siderealTimeInterval);
	    }
	    
	    	    
	    
	    // create a binary airmass	    
	    double [] binaryAirmass = new double[airmass.length];
	    for (int i = 0; i < airmass.length; i++) {	    	
	    	double airm = airmass[i];
	    	if ( airm < AirMass.NOT_VIS  && airm < airmassOBConstraint )
	    		binaryAirmass[i] = 1;
	    	else
	    		binaryAirmass[i] = 0;
		}
	    res = VisibilityCalcUtils.conjunction(res,binaryAirmass);
	    	    
	    
	    // iterate over the binary result, and replace 1 for the correct airmass
	    // and replace 0 for AirMass.NOT_VIS
	    for (int i = 0; i < res.length; i++) {
	    	
	    	if (res[i] == 1) {	    		
	    		if (airmass[i] < AirMass.NOT_VIS ) {
	    			res[i] = airmass[i];	    	
	    		}
	    	}	else {
	    		res[i] = AirMass.NOT_VIS;
	    	}
		}	
	    
	    // zenith 
	    res = confineZenithConstraint(interval,meetsZenithConstraint,options, res);	    
	    

	    return (res);
	}




	/**
	 * confine with the meet zenith constraint array
	 * the zenith constraint array has bigger precision (smaller step) than
	 * the airmass array
	 * 
	 * @param interval
	 * @param meetsZenithConstraint
	 * @param options
	 * @param res
	 */
	private double [] confineZenithConstraint(int interval, double[] meetsZenithConstraint, VisicalcOptions options, double[] res) {
			    
		double [] result = res;
		
	    int precisionZenith = options.getZenithConstraintCalcPrecision();
	    
	    int precisionAirmass = interval;	    	    
	    int precisionFactor;
	    
	    if ( precisionAirmass % precisionZenith == 0) {
	    	precisionFactor = precisionAirmass / precisionZenith;
	    } else {
	    	precisionZenith = 1;
	    	precisionFactor = precisionAirmass;
	    }
	    
	    
	    // determines if start and end indexes when the target crosses zenith
	    boolean crossedZenith = false;
	    int zenithStartIndex = 0;
	    int zenithEndIndex = 0;
	    
	    for (int i = 0; i < meetsZenithConstraint.length; i++) {
			if ( meetsZenithConstraint[i] == 0) {
				if (!crossedZenith) {
					crossedZenith = true;
					zenithStartIndex = i;
				}
			} else {
				if (crossedZenith) {
					zenithEndIndex = i;
					break;
				}
			}
		}
	    
	    if ( crossedZenith ) {	    
	    	//	convert the indexes to the precision of the airmass array
	    	for (int i = zenithStartIndex/precisionFactor; i <= zenithEndIndex/precisionFactor; i++) {
	    		result[i] = AirMass.NOT_VIS;
			}
	    
	    
	    }
	    
	    return result;
	}
	
	
	
	/**
	 * confine with the meet zenith constraint array
	 * the zenith constraint array has bigger precision (smaller step) than
	 * the airmass array
	 * 
	 * @param interval
	 * @param meetsZenithConstraint
	 * @param options
	 * @param res
	 */
	private double[] confineZenithConstraint(double[] lsts, double startLST, double endLST, double[] res) {

		double[] result = res;

		
		for (int i = 0; i < result.length; i++) {

			if (startLST <= lsts[i] && lsts[i] <= endLST)
				result[i] = AirMass.NOT_VIS;

		}

		return result;
	}
	
	
	
	/**
	 * 
	 * Dummy implementation of a sun rise time algorithm It simple iterates
	 * (using a stepInterval) until it finds a sun angle bigger than riseAngle
	 * 
	 * @param riseAngle
	 *            in degrees
	 * @param stepInterval
	 *            in minutes
	 * @param start
	 *            when to start calculating
	 * @param longit
	 *            longitude of observatory in radians
	 * @param lat
	 *            latitude of observatory in radians
	 * @param lst
	 *            local sidereal time of observatory
	 * @return
	 */
	public Date getSunRiseTime( 
			double riseAngle , 
			int stepInterval,  
			Date start ,  
			double longit, 
			double lat) {
			
					
		Date d = new Date(start.getTime());
								
		
		double jdate = JulianDate.getJDC(d);
		double lst = getLST(jdate,longit);
		double sunAngle = getSunValue(jdate,longit,lat,lst);
						
		boolean riseNext = true;
		if ( sunAngle > riseAngle )
			riseNext = false;
		
		

		while ( sunAngle < riseAngle || riseNext == false ) {
			d = new Date(d.getTime() + VisibilityCalcUtils.MILLISECS_PER_SEC*60*stepInterval  );  
			jdate = JulianDate.getJDC(d);				
			sunAngle = getSunValue(jdate,longit,lat,0);			

			if (sunAngle < riseAngle ) 
				riseNext  = true;
		}
			
		
		return(d);		
	}
	
	

}
